#%% ----------------------------------------------------------------------------
# INITIALISATION OF PYTHON e.g. packages, etc.
# ------------------------------------------------------------------------------

import numpy as np
import pandas as pd
import os
import sys
from pandas import read_csv
from pandas import read_excel
import math
import datetime
from datetime import timedelta

path = os.getcwd()



class Organism:
    ''' 
    Placeholder class which will later be replaced by the QSAR functionality of AquaPriori ?!

    For removal of microbial organisms ('mbo') / pathogen 
    
    microbial_species_dict: dict
    Attributes
    ---------
    organism_name: String
        species_name of the substance (for now limited dictionary to 'MS2' (MS2-virus)
        
    'alpha0': float
        reference_collision_efficiency [-]
        per redox zone ('suboxic', 'anoxic', deeply_anoxic')
    'reference_pH': float
        reference pH for calculating collision efficiency [-]
        per redox zone ('suboxic', 'anoxic', deeply_anoxic')
    'organism_diam': float
        diameter of pathogen/species [m]
    'mu1': float
        inactivation coefficient [1/day]
        per redox zone ('suboxic', 'anoxic', deeply_anoxic')
    '''  
    def __init__(self, organism_name, 
                    removal_function = 'mbo'):
        """
        Parameters
        ----------
        organism: String,
            name of the substance (for now limited dictionary to 'benzene', 'AMPA', 'benzo(a)pyrene'

        Returns
        -------
        organism_dict: dictionary
            'alpha0': float
                reference_collision_efficiency [-]
                per redox zone ('suboxic', 'anoxic', deeply_anoxic')
            'reference_pH': float
                reference pH for calculating collision efficiency [-]
                per redox zone ('suboxic', 'anoxic', deeply_anoxic')
            'organism_diam': float
                diameter of pathogen/species [m]
            'mu1': float
                inactivation coefficient [1/day]
                per redox zone ('suboxic', 'anoxic', deeply_anoxic')

        @SR (21-1-2022): Adjust documentation for microbial organisms (mbo)        
        """
        self.organism_name = organism_name

        # Dict    # Acties Steven 25-1-22

        # Naming convention organism: Uppercamelcase species
        micro_organism_dict = {
            "MS2": 
                {"organism_name": "MS2",
                    "alpha0": {
                        "suboxic": 1.e-3, 
                        "anoxic": 1.e-5, 
                        "deeply_anoxic": 1.e-5
                    },
                    "reference_pH": {
                        "suboxic": 6.6, 
                        "anoxic": 6.8, 
                        "deeply_anoxic": 6.8
                    },
                    "organism_diam": 2.33e-8,
                    "mu1": {
                        "suboxic": 0.039, 
                        "anoxic": 0.023, 
                        "deeply_anoxic": 0.023
                    }
                },
                # ALLES X 10
            "Escherichia coli": 
                {"organism_name": "Escherichia coli",
                    "alpha0": {
                        "suboxic": 1.e-2, 
                        "anoxic": 1.e-4, 
                        "deeply_anoxic": 1.e-4
                    },
                    "reference_pH": {
                        "suboxic": 6.6, 
                        "anoxic": 6.8, 
                        "deeply_anoxic": 6.8
                    },
                    "organism_diam": 1.5e-6,
                    "mu1": {
                        "suboxic": 0.39, 
                        "anoxic": 0.23, 
                        "deeply_anoxic": 0.23
                    }
                }
            }
        #@ Steven voeg toe: micro_organism_dict
        # self.micro_organism_dict = micro_organism_dict[substance_name]
        self.organism_dict = micro_organism_dict[self.organism_name]


class MicrobialRemoval():
    '''
    Returns concentration for a given microbial species.


    organism: object
        The microbial organism object with the microbial organism (mbo) of interest

        organism_dict: dictionary
            'alpha0': float
                reference_collision_efficiency [-]
                per redox zone ('suboxic', 'anoxic', deeply_anoxic')
            'reference_pH': float
                reference pH for calculating collision efficiency [-]
                per redox zone ('suboxic', 'anoxic', deeply_anoxic')
            'organism_diam': float
                diameter of pathogen/species [m]
            'mu1': float
                inactivation coefficient [1/day]
                per redox zone ('suboxic', 'anoxic', deeply_anoxic')
    
    '''

    def __init__(self,
                organism: Organism = 'MS2',
                partition_coefficient_water_organic_carbon=None,
                dissociation_constant=None,
                halflife_suboxic=None,
                halflife_anoxic=None,
                halflife_deeply_anoxic=None,
                alpha0_suboxic=None,
                alpha0_anoxic=None,
                alpha0_deeply_anoxic=None,
                reference_pH_suboxic=None,
                reference_pH_anoxic=None,
                reference_pH_deeply_anoxic=None,
                mu1_suboxic=None,
                mu1_anoxic=None,
                mu1_deeply_anoxic=None,
                organism_diam=None,
                ):

        '''
        Initialization of the MicrobialRemoval class, checks for user-defined 
        microbial organism removal parameters and overrides the database values.
        '''

        # Organism
        self.organism_name = organism
        self.alpha0_suboxic=alpha0_suboxic
        self.alpha0_anoxic=alpha0_anoxic
        self.alpha0_deeply_anoxic=alpha0_deeply_anoxic
        self.reference_pH_suboxic=reference_pH_suboxic
        self.reference_pH_anoxic=reference_pH_anoxic
        self.reference_pH_deeply_anoxic=reference_pH_deeply_anoxic
        self.mu1_suboxic=mu1_suboxic
        self.mu1_anoxic=mu1_anoxic
        self.mu1_deeply_anoxic=mu1_deeply_anoxic
        self.organism_diam=organism_diam        

        # Load (default) microbial organism data
        self.organism = Organism(organism_name = organism)

        # Create user dict with 'removal_parameters' from input
        # if self.removal_function == 'omp':
        user_removal_parameters = {
            self.organism_name:
                {"organism_name": self.organism_name,
                    "alpha0": {
                        "suboxic": self.alpha0_suboxic, 
                        "anoxic": self.alpha0_anoxic, 
                        "deeply_anoxic": self.alpha0_deeply_anoxic
                    },
                    "reference_pH": {
                        "suboxic": self.reference_pH_suboxic, 
                        "anoxic": self.reference_pH_anoxic, 
                        "deeply_anoxic": self.reference_pH_deeply_anoxic
                    },
                    "organism_diam": self.organism_diam,
                    "mu1": {
                        "suboxic": self.mu1_suboxic, 
                        "anoxic": self.mu1_anoxic, 
                        "deeply_anoxic": self.mu1_deeply_anoxic
                    }
                },
            }

        
        # User defined removal parameters [omp]
        user_removal_parameters = user_removal_parameters[self.organism_name]
        # assumes that default dict contains microbial organism input (only MS2 currently supported)
        default_removal_parameters = self.organism.organism_dict

        # iterate through the dictionary keys
        for key, value in user_removal_parameters.items():
            if type(value) is dict:
                for tkey, cvalue in value.items():
                    if cvalue is None: #reassign the value from the default dict if not input by the user
                        user_removal_parameters[key][tkey] = default_removal_parameters[key][tkey]
                    elif type(cvalue) is dict:
                        for subkey, subval in cvalue.items():
                            if subval is None:
                                user_removal_parameters[key][tkey][subkey] = default_removal_parameters[key][tkey][subkey]
                    # else: no assignment from default dict required...
            else:
                if value is None:
                    user_removal_parameters[key] = default_removal_parameters[key]
            
        #assign updated dict as attribute of the class to be able to access later
        self.removal_parameters = user_removal_parameters
        
    def calc_lambda(self, redox = 'anoxic',
                mu1 = 0.149, mu1_std = 0.0932,
                por_eff = 0.33,
                grainsize = 0.00025,
                pH = 7.5,
                temp_water = 10.,
                rho_water = 999.703,
                alpha0 = 0.001,
                reference_pH = 7.5,
                organism_diam = 2.33e-8, v_por = 0.01):
        
        ''' For more information about the advective microbial removal calculation: 
                BTO2012.015: Ch 6.7 (page 71-74)

        Calculate removal coefficient 'lambda_' [/day].

            lambda_ = k_att + mu_1
            with 'hechtingssnelheidscoëfficiënt k_att [/day] and
            inactivation constant 'mu1' [/day] - sub(oxic): 0.149 [/day]
            lognormal standarddev. 'mu1_std' - sub(oxic): 0.0932 [/day]

        First, calculate "hechtingssnelheidscoëfficiënt" 'k_att' [/day]
            for Particle Paths with id_vals 'part_idx'
        # Effective porosity ('por_eff') [-]
        # grain size 'grainsize' [m]
        # collision efficiency ('botsingsefficiëntie') 'bots_eff' [-]
        # Boltzmann constant (const_BM) [1,38 × 10-23 J K-1] 
        # Water temperature 'temp_water' [degrees celcius]
        # Water density 'rho_water' [kg m-3]
        # Organism/species diameter 'organism_diam' [m]
        # porewater velocity 'v_por' [m/d]
        
        # Check - (example 'E_coli'):
        >> k_att = calc_katt(part_idx = [0], por_eff = [0.33], korrelgrootte = [0.00025],
                bots_eff = 0.001, const_BM = 1.38e-23,
                temp_water = [10.], rho_water = [999.703],
                organism_diam = 2.33e-8, v_por = [0.01])
        >> print(k_att)
        >> {0: 0.7993188853572424} # [/day]
        '''

        # Boltzmann coefficient [J K-1]
        const_BM = 1.38e-23    

        # Collision efficiency
        coll_eff = alpha0 * 0.9**((pH - reference_pH)/0.1)

        # Collision term 'k_coll'
        k_coll = (3/2.)*((1-por_eff) / grainsize) * coll_eff

        # Porosity dependent variable 'gamma'
        gamma = (1-por_eff)**(1/3)

        # Calculate Happel’s porosity dependent parameter 'A_s' (Eq. 5: BTO2012.015)
        ''' !!! Use correct formula:-> As =  2 * (1-gamma**5) /  (2 - 3 * gamma + 3 * gamma**5 - 2 * gamma**6)
            instead of... 2 * (1-gamma)**5 / (.......) 
        '''
        As_happ = 2 * (1-gamma**5) / \
                (2 - 3 * gamma + 3 * gamma**5 - 2 * gamma**6)

        # Dynamic viscosity (mu) [kg m-1 s-1]
        mu = (rho_water * 497.e-6) / \
                    (temp_water + 42.5)**(3/2)


        # Diffusion constant 'D_BM' (Eq.6: BTO2012.015) --> unit: [m2 s-1]
        D_BM = (const_BM * (temp_water + 273.)) / \
                    (3 * np.pi * organism_diam * mu)
        # Diffusieconstante 'D_BM' (Eq.6: BTO2012.015) --> unit: [m2 d-1]
        D_BM *= 86400.

        # Diffusion related attachment term 'k_diff'
        k_diff = ((D_BM /
                    (grainsize * por_eff * v_por))**(2/3) * v_por)

        # hechtingssnelheidscoëfficiënt k_att [/dag] :'attachment coefficient related to porewater velocity'
        k_att = k_coll * 4 * As_happ**(1/3) * k_diff
        # removal coefficient 'lambda_' [/day], using the 'mu1' mean.
        lambda_ = k_att + mu1

        return lambda_, k_att
    
    def calc_advective_microbial_removal(self, grainsize = 0.00025,
                                        temp_water = 11., rho_water = 999.703,
                                        pH = 7.5, por_eff = 0.33,
                                        conc_start = 1., conc_gw = 0.,
                                        redox = 'anoxic',
                                        distance_traveled = 1., traveltime = 100.,
                                        mu1 = None, alpha0 = None, reference_pH = None,
                                        organism_diam = None):
        ''' Calculate the advective microbial removal of microbial organisms
            from source to end_point.

            For more information about the advective microbial removal calculation: 
                BTO2012.015: Ch 6.7 (page 71-74)

            Calculate the steady state concentration along traveled distance per 
            node for each pathline from startpoint to endpoint_id'.
            
            With removal coefficient 'lambda_' [/d] (redox dependent)
            effective porosity 'porosity' [-]
            Starting concentration 'conc_start' per pathline
            Initial groundwater concentration 'conc_gw'
            Distance between points 'distance_traveled' [m]
            Time between start and endpoint 'traveltime' [days]

            
        '''

        # mu1 [day -1]
        if mu1 is None:
            mu1 = self.removal_parameters['mu1'][redox]

        # alpha0 [-]
        if alpha0 is None:
            alpha0 = self.removal_parameters['alpha0'][redox]

        # reference pH [-]
        if reference_pH is None:
            reference_pH = self.removal_parameters['reference_pH'][redox]

        # organism diameter [m]
        if organism_diam is None:
            organism_diam = self.removal_parameters['organism_diam']

        # porewater_velocity
        v_por = distance_traveled / traveltime

        # Calculate removal coefficient lambda [day -1]
        self.lambda_, self.k_att = self.calc_lambda(redox = redox, mu1 = mu1,
                                    por_eff = por_eff, grainsize = grainsize, 
                                    pH = pH, 
                                    temp_water = temp_water, 
                                    rho_water = rho_water,
                                    alpha0 = alpha0, 
                                    reference_pH = reference_pH,
                                    organism_diam = organism_diam)

        # Calculate concentration after microbial removal in subsurface
        C_final = (conc_start - conc_gw) * np.exp(-(self.lambda_/v_por)*distance_traveled) + conc_gw


        # return final concentration 'C_final'
        return C_final






from dis import dis
import pytest
import numpy as np
import pandas as pd
import os
import sys
from pathlib import Path
from pandas.testing import assert_frame_equal
import warnings

from zmq import zmq_version_info

import sutra.sutra_functions as SF

# get directory of this file
path = Path(__file__).parent #os.getcwd() #path of working directory


#%%
def test_scenarios_mbo_removal_function(organism_name = "MS2"):

    # Location of test file for scenarios
    scenarios_fpath = os.path.join(path,"Testberekeningen_sutra_mbo_removal_220321.xlsx")
    sheet_name = "Scenarios"
    # Read scenario excel file
    df_test = pd.read_excel(scenarios_fpath, sheet_name = sheet_name, skiprows = 1)

    # df_output
    columns_output = ["k_att","lambda","steady_state_concentration"]
    df_output = pd.DataFrame(index = df_test.index, columns = columns_output)

    # Calculate advective microbial removal
    mbo_removal = SF.MicrobialRemoval(organism = organism_name)

    for fid in df_test.index:
        organism_name = "MS2"
        redox = df_test.at[fid,'redox']
        alpha0 = df_test.at[fid,'alpha0']
        reference_pH = df_test.at[fid,'reference_pH']
        mu1 = df_test.at[fid,'mu1']
        organism_diam = df_test.at[fid,'organism_diam']
        por_eff = df_test.at[fid,'porosity']
        grainsize = df_test.at[fid,'grainsize']
        pH_water = df_test.at[fid,'pH']
        temp_water = df_test.at[fid,'temperature']
        rho_water = df_test.at[fid,'rho_water']
        conc_start = 1.  # normally in df_flowline; use relative concentration as output
        conc_gw = 0.     # normally in df_flowline; use relative concentration as output
        distance_traveled = df_test.at[fid,'relative_distance']
        traveltime = df_test.at[fid,'total_travel_time']
 
        # Calculate final concentration after advective microbial removal
        C_final = mbo_removal.calc_advective_microbial_removal(grainsize = grainsize,
                                                temp_water = temp_water, rho_water = rho_water,
                                                pH = pH_water, por_eff = por_eff, 
                                                conc_start = conc_start, conc_gw = conc_gw,
                                                redox = redox,
                                                distance_traveled = distance_traveled, 
                                                traveltime = traveltime,
                                                organism_diam = organism_diam,
                                                mu1 = mu1,
                                                alpha0 = alpha0,
                                                reference_pH = reference_pH )

        # k_att, calculated
        df_output.loc[fid,"k_att"] = mbo_removal.k_att
        # lambda, calculated
        df_output.loc[fid,"lambda"] = mbo_removal.lambda_
        # (relative) concentration, calculated
        df_output.loc[fid,"steady_state_concentration"] = C_final

    # Calculate the difference between test dataframe and generated dataframe
    diff_perc = np.abs((df_output.loc[:,columns_output].values - \
                        df_test.loc[:,columns_output].values) / \
                        df_test.loc[:,columns_output].values) * 100.



    assert not np.any(diff_perc > 0.5)
    #print("dataframe values differ to much: " + str(round(diff_perc.max(),2)) + " %")

#%%
def test_mbo_removal_function_check_default(organism_name = "MS2",
                                            redox = 'anoxic',
                                            alpha0 = 1.e-5,
                                            reference_pH = 6.8,
                                            mu1 = 0.023,
                                            organism_diam = 2.33e-8,
                                            por_eff = 0.33,
                                            grainsize = 0.00025,
                                            pH_water = 7.5,
                                            temp_water = 11.,
                                            rho_water = 999.703,
                                            conc_start = 1.,
                                            conc_gw = 0.,
                                            distance_traveled = 1.,
                                            traveltime = 100.):

    ## Default test
    # Calculate advective microbial removal
    mbo_removal_default = SF.MicrobialRemoval(organism = organism_name)
    # Calculate final concentration after advective microbial removal
    C_final_default= mbo_removal_default.calc_advective_microbial_removal()

    # Lambda (default): inactivation 
    lambda_default = mbo_removal_default.lambda_   

    '''
    ## Default parameters: ##
    redox = 'anoxic',
    alpha0 = 1.e-5,
    reference_pH = 6.8,
    mu1 = 0.023,
    organism_diam = 2.33e-8,
    por_eff = 0.33,
    grainsize = 0.00025,
    pH_water = 7.5,
    temp_water = 11.,
    rho_water = 999.703,
    conc_start = 1.,
    conc_gw = 0.,
    distance_traveled = 1.,
    traveltime = 100.
    '''

    # Calculate advective microbial removal
    mbo_removal_test = SF.MicrobialRemoval(organism = organism_name)
    # Calculate final concentration after advective microbial removal
    C_final_test = mbo_removal_test.calc_advective_microbial_removal(grainsize = grainsize,
                                            temp_water = temp_water, rho_water = rho_water,
                                            pH = pH_water, por_eff = por_eff, 
                                            conc_start = 1., conc_gw = 0.,
                                            redox = 'anoxic',
                                            distance_traveled = distance_traveled, 
                                            traveltime = traveltime,
                                            organism_diam = organism_diam,
                                            mu1 = mu1,
                                            alpha0 = alpha0,
                                            reference_pH = reference_pH
                                            )

    # Lambda: inactivation 
    lambda_test = mbo_removal_test.lambda_     

    assert round(lambda_default,4) == round(lambda_test,4) 
    assert round(C_final_default,4) == round(C_final_test,4)



def test_manual_input_mbo_removal(organism_name = "MS2"):

    # test parameters
    organism_name = organism_name
    por_eff = 0.33
    grainsize = 0.00025
    pH_water = 7.5
    temp_water = 10.
    rho_water = 999.703

    # Removal parms
    # alpha  'sticky coefficient'
    alpha0 = 0.001 # [-]
    reference_pH = 7.5
    # --> if pH == reference_pH, then coll_eff == alpha0
    # coll_eff = 0.001

    # time dependent inactivation coefficient mu1 [day-1]
    mu1 = 0.149
    # org. diameter [m]
    organism_diam = 2.33e-8

    distance_traveled = 1.
    traveltime = 100.
    porewater_velocity = distance_traveled / traveltime

    # Calculate advective microbial removal
    mbo_removal = SF.MicrobialRemoval(organism = organism_name)
    # Calculate advective microbial removal
    C_final = mbo_removal.calc_advective_microbial_removal(grainsize = grainsize,
                                            temp_water = temp_water, rho_water = rho_water,
                                            pH = pH_water, por_eff = por_eff, 
                                            conc_start = 1., conc_gw = 0.,
                                            redox = 'anoxic',
                                            distance_traveled = distance_traveled, 
                                            traveltime = traveltime,
                                            organism_diam = organism_diam,
                                            mu1 = mu1,
                                            alpha0 = alpha0,
                                            reference_pH = reference_pH
                                            )
    # Lambda: inactivation 
    lambda_ = mbo_removal.lambda_                                             

    assert round(lambda_,4) == round(0.7993188853572424 + mu1,4) 

    assert round(C_final,3) == round(6.531818379725895e-42,3)
    


"""  Convenience file to easily get the full path of the project. """
import sys
from pathlib import Path
import os

module_path = Path(__file__).parent.parent
if module_path not in sys.path:
    sys.path.append(str(module_path))

